import { CryptoUtil } from './crypto.js';
import { ReputationHolder } from './reputation-holder.js';
import { Stake } from './stake.js';
import { Voter } from './voter.js';
import params from '../params.js';

const ValidationPoolStates = Object.freeze({
  OPEN: 'OPEN',
  CLOSED: 'CLOSED',
  RESOLVED: 'RESOLVED',
});

/**
 * Purpose: Enable voting
 */
export class ValidationPool extends ReputationHolder {
  constructor(
    bench,
    forum,
    {
      postId,
      reputationPublicKey,
      fee,
      duration,
      tokenLossRatio,
      contentiousDebate = false,
      authorStakeAmount = 0,
    },
    name,
    scene,
  ) {
    super(`pool_${CryptoUtil.randomUUID()}`, name, scene);
    this.id = this.reputationPublicKey;
    // If contentiousDebate = true, we will follow the progression defined by getTokenLossRatio()
    if (
      !contentiousDebate
      && (tokenLossRatio < 0
        || tokenLossRatio > 1
        || [null, undefined].includes(tokenLossRatio))
    ) {
      throw new Error(
        `Token loss ratio must be in the range [0, 1]; got ${tokenLossRatio}`,
      );
    }
    if (
      duration < params.voteDuration.min
      || (params.voteDuration.max && duration > params.voteDuration.max)
      || [null, undefined].includes(duration)
    ) {
      throw new Error(
        `Duration must be in the range [${params.voteDuration.min}, ${
          params.voteDuration.max ?? 'Inf'
        }]; got ${duration}`,
      );
    }
    this.bench = bench;
    this.forum = forum;
    this.postId = postId;
    this.state = ValidationPoolStates.OPEN;
    this.setStatus('Open');
    this.stakes = new Set();
    this.dateStart = new Date();
    this.authorReputationPublicKey = reputationPublicKey;
    this.fee = fee;
    this.duration = duration;
    this.tokenLossRatio = tokenLossRatio;
    this.contentiousDebate = contentiousDebate;
    this.mintedValue = fee * params.mintingRatio();
    this.tokenId = this.bench.reputation.mint(this.id, this.mintedValue);
    // Tokens minted "for" the post go toward stake of author voting for their own post.
    // Also, author can provide additional stakes, e.g. availability stakes for work evidence post.
    this.stake(this, {
      position: true,
      amount: this.mintedValue * params.stakeForAuthor + authorStakeAmount,
      tokenId: this.tokenId,
    });
    this.stake(this, {
      position: false,
      amount: this.mintedValue * (1 - params.stakeForAuthor),
      tokenId: this.tokenId,
    });
  }

  getTokenLossRatio() {
    if (!this.contentiousDebate) {
      return this.tokenLossRatio;
    }
    const elapsed = new Date() - this.dateStart;
    let stageDuration = params.contentiousDebate.period / 2;
    let stage = 0;
    let t = 0;
    while (true) {
      t += stageDuration;
      stageDuration /= 2;
      if (t > elapsed) {
        break;
      }
      stage += 1;
      if (stage >= params.contentiousDebate.stages - 1) {
        break;
      }
    }
    return stage / (params.contentiousDebate.stages - 1);
  }

  /**
   * @param {boolean} outcome: null --> all entries. Otherwise filters to position === outcome.
   * @param {boolean} options.excludeSystem: Whether to exclude votes cast during pool initialization
   * @returns stake[]
   */
  getStakes(outcome, { excludeSystem }) {
    return Array.from(this.stakes.values())
      .filter(({ tokenId }) => !excludeSystem || tokenId !== this.tokenId)
      .filter(({ position }) => outcome === null || position === outcome);
  }

  /**
   * @param {boolean} outcome: null --> all entries. Otherwise filters to position === outcome.
   * @returns number
   */
  getTotalStakedOnPost(outcome) {
    return this.getStakes(outcome, { excludeSystem: false })
      .map((stake) => stake.getStakeValue())
      .reduce((acc, cur) => (acc += cur), 0);
  }

  /**
   * @param {boolean} outcome: null --> all entries. Otherwise filters to position === outcome.
   * @returns number
   */
  getTotalValueOfStakesForOutcome(outcome) {
    return this.getStakes(outcome, { excludeSystem: false })
      .reduce((total, { amount }) => (total += amount), 0);
  }

  // TODO: This can be handled as a hook on receipt of reputation token transfer
  async stake(reputationHolder, {
    tokenId, position, amount, lockingTime = 0,
  }) {
    if (this.state === ValidationPoolStates.CLOSED) {
      throw new Error(`Validation pool ${this.id} is closed.`);
    }

    if (this.duration && new Date() - this.dateStart > this.duration) {
      throw new Error(
        `Validation pool ${this.id} has expired, no new votes may be cast.`,
      );
    }

    const { reputationPublicKey } = reputationHolder;
    if (reputationPublicKey !== this.bench.reputation.ownerOf(tokenId)) {
      throw new Error('Reputation may only be staked by its owner!');
    }

    const stake = new Stake({
      tokenId, position, amount, lockingTime,
    });
    this.stakes.add(stake);

    // Transfer staked amount from the sender to the validation pool
    this.bench.reputation.transferValueFrom(tokenId, this.tokenId, amount);

    // Keep a record of voters and their votes
    if (tokenId !== this.tokenId) {
      const voter = this.bench.voters.get(reputationPublicKey) ?? new Voter(reputationPublicKey);
      voter.addVoteRecord(this);
      this.bench.voters.set(reputationPublicKey, voter);
    }
  }

  applyTokenLocking() {
    // Before evaluating the winning conditions,
    // we need to make sure any staked tokens are locked for the
    // specified amounts of time.
    for (const { tokenId, amount, lockingTime } of this.stakes.values()) {
      this.bench.reputation.lock(tokenId, amount, lockingTime);
      // TODO: If there is an exception here, the voter may have voted incorrectly. Consider penalties.
    }
  }

  async evaluateWinningConditions() {
    if (this.state === ValidationPoolStates.RESOLVED) {
      throw new Error('Validation pool has already been resolved!');
    }
    const elapsed = new Date() - this.dateStart;
    if (elapsed < this.duration) {
      throw new Error(`Validation pool duration has not yet elapsed! ${this.duration - elapsed} ms remaining.`);
    }
    // Now we can evaluate winning conditions
    this.state = ValidationPoolStates.CLOSED;
    this.setStatus('Closed');

    const upvoteValue = this.getTotalValueOfStakesForOutcome(true);
    const downvoteValue = this.getTotalValueOfStakesForOutcome(false);
    const activeAvailableReputation = this.bench.getActiveAvailableReputation();
    const votePasses = upvoteValue >= params.winningRatio * downvoteValue;
    const quorumMet = upvoteValue + downvoteValue >= params.quorum * activeAvailableReputation;

    const result = {
      votePasses,
      upvoteValue,
      downvoteValue,
    };

    if (quorumMet) {
      this.setStatus(`Resolved - ${votePasses ? 'Won' : 'Lost'}`);
      this.scene.sequence.log(`note over ${this.name} : ${votePasses ? 'Win' : 'Lose'}`);
      this.applyTokenLocking();
      await this.distributeReputation({ votePasses });
      // TODO: distribute fees
    } else {
      this.setStatus('Resolved - Quorum not met');
      this.scene.sequence.log(`note over ${this.name} : Quorum not met`);
    }

    this.deactivate();
    this.state = ValidationPoolStates.RESOLVED;
    return result;
  }

  async distributeReputation({ votePasses }) {
    // For now we assume a tightly binding pool, where all staked reputation is lost
    // TODO: Take tokenLossRatio into account
    // TODO: revoke staked reputation from losing voters

    // In a tightly binding validation pool, losing voter stakes are transferred to winning voters.
    const tokensForWinners = this.getTotalStakedOnPost(!votePasses);
    const winningEntries = this.getStakes(votePasses, { excludeSystem: true });
    const totalValueOfStakesForWin = this.getTotalValueOfStakesForOutcome(votePasses);

    // Compute rewards for the winning voters, in proportion to the value of their stakes.
    for (const stake of winningEntries) {
      const { tokenId, amount } = stake;
      const value = stake.getStakeValue();
      const reward = tokensForWinners * (value / totalValueOfStakesForWin);
      // Also return each winning voter their staked amount
      const reputationPublicKey = this.bench.reputation.ownerOf(tokenId);
      console.log(`reward for winning stake by ${reputationPublicKey}: ${reward}`);
      this.bench.reputation.transferValueFrom(this.tokenId, tokenId, reward + amount);
    }

    if (votePasses && !!this.forum) {
      // Distribute awards to author via the forum
      // const tokensForAuthor = this.mintedValue * params.stakeForAuthor + rewards.get(this.tokenId);
      console.log(`sending reward for author stake to forum: ${this.bench.reputation.valueOf(this.tokenId)}`);

      // Transfer ownership of the minted token, from the pool to the forum
      this.bench.reputation.transferFrom(this.id, this.forum.id, this.tokenId);

      // Recurse through forum to determine reputation effects
      await this.forum.onValidate({
        bench: this.bench,
        pool: this,
        postId: this.postId,
        tokenId: this.tokenId,
      });
    }

    console.log('pool complete');
  }
}
